_prefix=/usr
_bindir=/usr/bin

cd ppd
./autogen.sh --prefix=${_prefix} 

cd ../pstoufr2cpca
./autogen.sh --prefix=${_prefix} --enable-progpath=${_bindir} --enable-shared --disable-static

cd ..
make

